// German's Cache Coherence Protocol

/*Check safety like this:
  ./ccp 5 -spec "Bad & (FS Init)" -iter
  Or:
  ./ccp 3 -spec "Init & (EF Bad)" -iter
  Check communal progress like this:
  ./ccp 5 -spec "(FS Init) & req_ex_i & EG (! gr_ex_i)" -iter
*/

Parameter
  n ("number of processes");

Order
  concat;
  // interl;

Global
  heg: bool;
  hcm: enum { null, req_sh, req_ex };
  hci: SB(0);

Clique[n]
  c: enum { I, S, E };
  hsl: bool;
  hil: bool;
  ch1: enum { null, req_sh, req_ex };
  ch2: enum { null, inv, gr_sh, gr_ex };
  ch3: enum { null, inv_ack };

Proposition
  Init;
  Bad;
  req_ex_i;
  gr_ex_i;

Code

UBDD rule01(const StateSpace& S) {
  UBDD rule01 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.c  (p), StateSpace::  c::I);
    for_p &= UBDD::equal(S.ch1(p), StateSpace::ch1::null);

    for_p &= UBDD::equal(S._ch1(p), StateSpace::ch1::req_sh);

    for_p &= UBDD::invariant(S.c  (p));
    for_p &= UBDD::invariant(S.hsl(p));
    for_p &= UBDD::invariant(S.hil(p));
    for_p &= UBDD::invariant(S.ch2(p));
    for_p &= UBDD::invariant(S.ch3(p));

    rule01 |= for_p; }

  rule01 &= UBDD::invariant(S.heg());
  rule01 &= UBDD::invariant(S.hcm());
  rule01 &= UBDD::invariant(S.hci());

  return rule01; }

UBDD rule02(const StateSpace& S) {
  UBDD rule02 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.c  (p), StateSpace::  c::I);
    for_p &= UBDD::equal(S.ch1(p), StateSpace::ch1::null);

    for_p &= UBDD::equal(S._ch1(p), StateSpace::ch1::req_ex);

    for_p &= UBDD::invariant(S.c  (p));
    for_p &= UBDD::invariant(S.hsl(p));
    for_p &= UBDD::invariant(S.hil(p));
    for_p &= UBDD::invariant(S.ch2(p));
    for_p &= UBDD::invariant(S.ch3(p));

    rule02 |= for_p; }

  rule02 &= UBDD::invariant(S.heg());
  rule02 &= UBDD::invariant(S.hcm());
  rule02 &= UBDD::invariant(S.hci());

  return rule02; }

UBDD rule03(const StateSpace& S) {
  UBDD rule03 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = UBDD::One();

    for_p &= ! UBDD::equal( S.ch1(p), StateSpace::ch1::null);

    for_p &=   UBDD::equal(S._hcm(), S.ch1(p));
    for_p &=   UBDD::equal(S._ch1(p), StateSpace::ch1::null);
    for_p &=   UBDD::equal(S._hci(), p);

    for_p &=   UBDD::equal(S._hil(p), S.hsl(p));

    for_p &= UBDD::invariant(S.c  (p));
    for_p &= UBDD::invariant(S.hsl(p));
    for_p &= UBDD::invariant(S.ch2(p));
    for_p &= UBDD::invariant(S.ch3(p));

    for (ushort q = 0; q < S.n; q++)
      if (q != p) {
	for_p &= UBDD::equal(S._hil(q), S.hsl(q));
	for_p &= UBDD::invariant(S.c  (q));
	for_p &= UBDD::invariant(S.hsl(q));
	for_p &= UBDD::invariant(S.ch1(q));
	for_p &= UBDD::invariant(S.ch2(q));
	for_p &= UBDD::invariant(S.ch3(q)); }

    rule03 |= for_p; }

  rule03 &= UBDD::equal(S.hcm(), StateSpace::hcm::null);

  rule03 &= UBDD::invariant(S.heg());

  return rule03; }

UBDD rule04(const StateSpace& S) {
  UBDD rule04 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.hil(p), true);
    for_p &= UBDD::equal(S.ch2(p), StateSpace::ch2::null);

    for_p &= UBDD::equal(S._hil(p), false);
    for_p &= UBDD::equal(S._ch2(p), StateSpace::ch2::inv);

    for_p &= UBDD::invariant(S.c  (p));
    for_p &= UBDD::invariant(S.hsl(p));
    for_p &= UBDD::invariant(S.ch1(p));
    for_p &= UBDD::invariant(S.ch3(p));

    rule04 |= for_p; }

  rule04 &= (UBDD::equal(S.hcm(), StateSpace::hcm::req_sh) & UBDD::equal(S.heg(), true)) | UBDD::equal(S.hcm(), StateSpace::hcm::req_ex);

  rule04 &= UBDD::invariant(S.heg());
  rule04 &= UBDD::invariant(S.hcm());
  rule04 &= UBDD::invariant(S.hci());

  return rule04; }

UBDD rule05(const StateSpace& S) {
  UBDD rule05 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.ch3(p), StateSpace::ch3::inv_ack);

    for_p &= UBDD::equal(S._hsl(p), false);
    for_p &= UBDD::equal(S._ch3(p), StateSpace::ch3::null);

    for_p &= UBDD::invariant(S.c  (p));
    for_p &= UBDD::invariant(S.hil(p));
    for_p &= UBDD::invariant(S.ch1(p));
    for_p &= UBDD::invariant(S.ch2(p));

    rule05 |= for_p; }

  rule05 &= ! UBDD::equal(S.hcm(), StateSpace::hcm::null);

  rule05 &= UBDD::equal(S._heg(), false);

  rule05 &= UBDD::invariant(S.hcm());
  rule05 &= UBDD::invariant(S.hci());

  return rule05; }

UBDD rule06(const StateSpace& S) {
  UBDD rule06 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.ch2(p), StateSpace::ch2::inv);
    for_p &= UBDD::equal(S.ch3(p), StateSpace::ch3::null);

    for_p &= UBDD::equal(S._c  (p), StateSpace::  c::I);
    for_p &= UBDD::equal(S._ch2(p), StateSpace::ch2::null);
    for_p &= UBDD::equal(S._ch3(p), StateSpace::ch3::inv_ack);

    for_p &= UBDD::invariant(S.hsl(p));
    for_p &= UBDD::invariant(S.hil(p));
    for_p &= UBDD::invariant(S.ch1(p));

    rule06 |= for_p; }

  rule06 &= UBDD::invariant(S.heg());
  rule06 &= UBDD::invariant(S.hcm());
  rule06 &= UBDD::invariant(S.hci());

  return rule06; }

UBDD rule07(const StateSpace& S) {
  UBDD rule07 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.ch2(p), StateSpace::ch2::gr_sh);

    for_p &= UBDD::equal(S._c  (p), StateSpace::c  ::S);
    for_p &= UBDD::equal(S._ch2(p), StateSpace::ch2::null);

    for_p &= UBDD::invariant(S.hsl(p));
    for_p &= UBDD::invariant(S.hil(p));
    for_p &= UBDD::invariant(S.ch1(p));
    for_p &= UBDD::invariant(S.ch3(p));

    rule07 |= for_p; }

  rule07 &= UBDD::invariant(S.heg());
  rule07 &= UBDD::invariant(S.hcm());
  rule07 &= UBDD::invariant(S.hci());

  return rule07; }

UBDD rule08(const StateSpace& S) {
  UBDD rule08 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.ch2(p), StateSpace::ch2::gr_ex);

    for_p &= UBDD::equal(S._c  (p), StateSpace::c  ::E);
    for_p &= UBDD::equal(S._ch2(p), StateSpace::ch2::null);

    for_p &= UBDD::invariant(S.hsl(p));
    for_p &= UBDD::invariant(S.hil(p));
    for_p &= UBDD::invariant(S.ch1(p));
    for_p &= UBDD::invariant(S.ch3(p));

    rule08 |= for_p; }

  rule08 &= UBDD::invariant(S.heg());
  rule08 &= UBDD::invariant(S.hcm());
  rule08 &= UBDD::invariant(S.hci());

  return rule08; }

UBDD rule09(const StateSpace& S) {
  UBDD rule09 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.hci(), p);
    for_p &= UBDD::equal(S.ch2(p), StateSpace::ch2::null);

    for_p &= UBDD::equal(S._hsl(p), true);
    for_p &= UBDD::equal(S._ch2(p), StateSpace::ch2::gr_sh);

    for_p &= UBDD::invariant(S.c  (p));
    for_p &= UBDD::invariant(S.hil(p));
    for_p &= UBDD::invariant(S.ch1(p));
    for_p &= UBDD::invariant(S.ch3(p));

    rule09 |= for_p; }

  rule09 &= UBDD::equal( S.heg(), false);
  rule09 &= UBDD::equal( S.hcm(), StateSpace::hcm::req_sh);

  rule09 &= UBDD::equal(S._hcm(), StateSpace::hcm::null);

  rule09 &= UBDD::invariant(S.heg());
  rule09 &= UBDD::invariant(S.hci());

  return rule09; }

UBDD rule10(const StateSpace& S) {
  UBDD rule10 = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {
    UBDD for_p = S.invariant_but(p, 0);

    for_p &= UBDD::equal(S.hci(), p);
    for_p &= UBDD::equal(S.ch2(p), StateSpace::ch2::null);

    for_p &= UBDD::equal(S._hsl(p), true);
    for_p &= UBDD::equal(S._ch2(p), StateSpace::ch2::gr_ex);

    for_p &= UBDD::invariant(S.c  (p));
    for_p &= UBDD::invariant(S.hil(p));
    for_p &= UBDD::invariant(S.ch1(p));
    for_p &= UBDD::invariant(S.ch3(p));

    rule10 |= for_p; }

  for (ushort q = 0; q < S.n; q++)
    rule10 &= UBDD::equal(S.hsl(q), false);

  rule10 &= UBDD::equal(S.hcm(), StateSpace::hcm::req_ex);

  rule10 &= UBDD::equal(S._hcm(), StateSpace::hcm::null);
  rule10 &= UBDD::equal(S._heg(), true);

  rule10 &= UBDD::invariant(S.hci());

  return rule10; }

// Transition relation
UBDD R(const StateSpace& S) {
  UBDD R = UBDD::Zero();

  R |= rule01(S);
  R |= rule02(S);
  R |= rule03(S);
  R |= rule04(S);
  R |= rule05(S);
  R |= rule06(S);
  R |= rule07(S);
  R |= rule08(S);
  R |= rule09(S);
  R |= rule10(S);

  return R; }

// Create initial states
UBDD Init(const StateSpace& S) {
  UBDD Init = UBDD::One();

  Init &= UBDD::equal(S.heg(), false);
  Init &= UBDD::equal(S.hcm(), StateSpace::hcm::null);
  // hci is non-deterministic

  for (ushort p = 0; p < S.n; p++) {
    Init &= UBDD::equal(S.c  (p), StateSpace::c::I);
    Init &= UBDD::equal(S.hsl(p), false);
    Init &= UBDD::equal(S.hil(p), false);
    Init &= UBDD::equal(S.ch1(p), StateSpace::ch1::null);
    Init &= UBDD::equal(S.ch2(p), StateSpace::ch2::null);
    Init &= UBDD::equal(S.ch3(p), StateSpace::ch3::null); }

  return Init; }

// Returns the BDD for the set of bad states, i.e. where one process is in E and another one is not in I
UBDD Bad(const StateSpace& S) {
  UBDD Bad = UBDD::Zero();
  for (ushort p_i = 0; p_i < S.n; p_i++) {
    UBDD E = UBDD::equal(S.c(p_i), StateSpace::c::E);       // optimization: pulled out of the loop. This makes it a bit hard to read
    for (ushort p_j = 0; p_j < S.n; p_j++)
      if (p_i != p_j)
	Bad |= E & ! UBDD::equal(S.c(p_j), StateSpace::c::I); }
  return Bad; }

UBDD req_ex_i(const StateSpace& S) {
  UBDD result = UBDD::Zero();
  for (ushort p = 0; p < S.n; p++)
    result |= UBDD::equal(S.ch1(p), StateSpace::ch1::req_ex);
  return result; }

UBDD gr_ex_i(const StateSpace& S) {
  UBDD result = UBDD::Zero();
  for (ushort p = 0; p < S.n; p++)
    result |= UBDD::equal(S.c(p), StateSpace::c::E);
  return result; }
