// Token ring solution to Mutex

Parameter
  n ("number of processes");

Order
  // interl;
  concat;

Global
  tok: SB(0);

Ring[n]
  L: enum { N, T, C };

Proposition
  Init;
  Bad;

Code

// Transition relation
UBDD R(const StateSpace& S) {
  UBDD R = UBDD::Zero();

  for (ushort p = 0; p < S.n; p++) {

    UBDD for_p = UBDD::Zero();

    // N -> T
    for_p |= UBDD::equal(S.L(p), StateSpace::L::N) & UBDD::equal(S._L(p), StateSpace::L::T) & UBDD::invariant(S.tok());

    // T -> C
    for_p |= UBDD::equal(S.L(p), StateSpace::L::T) & UBDD::equal(S._L(p), StateSpace::L::C) & UBDD::invariant(S.tok()) & UBDD::equal(S.tok(), p);

    // C -> N
    for_p |= UBDD::equal(S.L(p), StateSpace::L::C) & UBDD::equal(S._L(p), StateSpace::L::N) & UBDD::equal(S._tok(), ( p == S.n - 1 ? 0 : p + 1 )); // ring symmetry
    // Replacing the previous line with the following and replacing the word "Ring" above with "Clique" gives a fully symmetric system
    // for_p |= UBDD::equal(S.L(p), StateSpace::L::C) & UBDD::equal(S._L(p), StateSpace::L::N) & UBDD::atMost(S._tok(), S.n - 1);

    R |= for_p & S.invariant_but(p, 0); }

  return R; }

// Create initial states
UBDD Init(const StateSpace& S) {
  UBDD Init = UBDD::One();
  // token is non-det within [0..n-1]
  for (ushort p = 0; p < S.n; p++)
    Init &= UBDD::equal(S.L(p), StateSpace::L::N);
  return Init; }

// Returns the BDD for the set of bad states, i.e. two processes in C
UBDD Bad(const StateSpace& S) {
  UBDD Bad = UBDD::Zero();

  for (ushort p_i = 0; p_i < S.n; p_i++)
    for (ushort p_j = 0; p_j < p_i; p_j++)
      if (p_i != p_j)
	Bad |= UBDD::equal(S.L(p_i), StateSpace::L::C) & UBDD::equal(S.L(p_j), StateSpace::L::C);

  return Bad; }
