// Readers-Writers protocol

Parameter
  r ("number of readers", "1");
  w ("number of writers", "1");

Const
  READERS = 0;  // symmetry block 0 represents the readers
  WRITERS = 1;  // symmetry block 1 represents the writers

Order
  concat;

Clique[r]
  rstate: enum { N, T, C };

Clique[w]
  wstate: enum { N, T, C };

Proposition
  Init;
  Bad;

Code

UBDD exists_reader_C(const StateSpace& S) {
  UBDD result = UBDD::Zero();
  for (ushort p = 0; p < S.r; p++)
    result |= UBDD::equal(S.rstate(p), StateSpace::rstate::C);
  return result; }

UBDD exists_writer_C(const StateSpace& S) {
  UBDD result = UBDD::Zero();
  for (ushort p = 0; p < S.w; p++)
    result |= UBDD::equal(S.wstate(p), StateSpace::wstate::C);
  return result; }

inline UBDD exists_C(const StateSpace& S) { return exists_reader_C(S) | exists_writer_C(S); }

UBDD R(const StateSpace& S) {
  UBDD R = UBDD::Zero();

  UBDD for_p;

  // Readers. Several of them are allowed in the critical section
  for (ushort p = 0; p < S.r; p++) {
    for_p = UBDD::Zero();
    // N --> T
    for_p |= UBDD::equal(S.rstate(p), StateSpace::rstate::N) & UBDD::equal(S._rstate(p),StateSpace::rstate::T);
    // T --> C
    for_p |= UBDD::equal(S.rstate(p), StateSpace::rstate::T) & UBDD::equal(S._rstate(p),StateSpace::rstate::C) & (! exists_writer_C(S));
    // C --> N
    for_p |= UBDD::equal(S.rstate(p), StateSpace::rstate::C) & UBDD::equal(S._rstate(p),StateSpace::rstate::N);

    for_p &= S.invariant_but(p, S.READERS) & S.invariant_sbn(S.WRITERS);

    R |= for_p; }

  // Writers. At most one of them may be in the critical section
  for (ushort p = 0; p < S.w; p++) {
    for_p = UBDD::Zero();
    // N --> T
    for_p |= UBDD::equal(S.wstate(p), StateSpace::wstate::N) & UBDD::equal(S._wstate(p),StateSpace::wstate::T);
    // T --> C
    for_p |= UBDD::equal(S.wstate(p), StateSpace::wstate::T) & UBDD::equal(S._wstate(p),StateSpace::wstate::C) & (! exists_C(S));
    // C --> N
    for_p |= UBDD::equal(S.wstate(p), StateSpace::wstate::C) & UBDD::equal(S._wstate(p),StateSpace::wstate::N);

    for_p &= S.invariant_but(p, S.WRITERS) & S.invariant_sbn(S.READERS);

    R |= for_p; }

  return R; }

UBDD Init(const StateSpace& S) {
  UBDD Init = UBDD::One();
  for (ushort p = 0; p < S.r; p++)
    Init &= UBDD::equal(S.rstate(p), StateSpace::rstate::N);
  for (ushort p = 0; p < S.w; p++)
    Init &= UBDD::equal(S.wstate(p), StateSpace::wstate::N);
  return Init; }

// Returns the BDD for the set of bad states: a writer + another process in C
UBDD Bad(const StateSpace& S) {
  UBDD Bad = UBDD::Zero();
  // two writers in C
  for (ushort p = 0; p < S.w; p++)
    for (ushort q = 0; q < S.w; q++)
      if (p != q)
	Bad |= UBDD::equal(S.wstate(p), StateSpace::rstate::C) & UBDD::equal(S.wstate(q), StateSpace::rstate::C);
  // a writer and a reader in C
  for (ushort p = 0; p < S.r; p++)
    for (ushort q = 0; q < S.w; q++)
      Bad |= UBDD::equal(S.rstate(p), StateSpace::rstate::C) & UBDD::equal(S.wstate(q), StateSpace::rstate::C);

  return Bad; }
