%{

// Tokenizing CTL formulas. "Atomic propositions" are C-style identifiers.

#include <stdlib.h> // for exit

#include <iostream>

#include "heap.hh"
#include "utool.hh"
#include "model.hh"
#include "model-check.hh"

#define YYSTYPE UBDD* // needed in ctl.y.h

typedef Heap<UBDD> UBDD_Heap;

extern UBDD_Heap ubdd_heap;

#include "ctl.y.h"

#define YY_DECL int yylex(ModelCheck* MC, Model& M) // non-const since we might build BDDs for APs here

extern UBDD* yylval; // yacc's global

%}

digit      [0-9]
letter     [A-Za-z]
identifier {letter}({letter}|{digit}|_)*
whitespace [ \t]

%%

{whitespace}+ ;

"EU"          return     EU_TOK;
"AU"          return     AU_TOK;

"EX"          return     EX_TOK;
"AX"          return     AX_TOK;
"EY"          return     EY_TOK;
"AY"          return     AY_TOK;
"EF"          return     EF_TOK;
"AF"          return     AF_TOK;
"EG"          return     EG_TOK;
"AG"          return     AG_TOK;
"EP"          return     EP_TOK;
"EX_INF"      return EX_INF_TOK;
"EY_INF"      return EY_INF_TOK;
"FS"          return     FS_TOK;
"BS"          return     BS_TOK;
"INV_FS"      return INV_FS_TOK;
"INV_BS"      return INV_BS_TOK;

"!" |
"|" |
"&"           return yytext[0];

"=="          return   EQUIV_TOK;
"=>"          return IMPLIES_TOK;

"False"       return FALSE_TOK;
"True"        return  TRUE_TOK;

"neg"         return NEG_TOK;

"States"      return STATES_TOK;
"R"           return R_TOK;
"size"        return SIZE_TOK;
"alpha"       return ALPHA_TOK;
"orbit"       return ORBIT_TOK;

{identifier}  { yylval = M.AP_pointer(yytext); return IDENTIFIER_TOK; }

"(" |
")"           return yytext[0];

.|\n          { std::cerr << "Token error: '" << yytext[0] << "'; exiting" << std::endl; exit(1); }

%%

// To avoid warning about unused yyunput:
int __YYUNPUT_DUMMY__() {
  yyunput(0, 0);
  return 0; }
