
#ifndef __HEAP_HH__
#define __HEAP_HH__

#include <set>
#include <string>
#include <iostream>

using std::set;
using std::string;

// A simple heap class. "request" returns a pointer to type T and adds it
// to the heap (parameterless constructor must exist). "insert" is used
// if the pointer already exists and only needs to be added to the
// heap. Use "erase" if you have deleted the pointer yourself and just need
// to remove it from the heap. "cleanup" frees all pointers and collapses
// the heap.  This class is useful if the precise momenent when an
// allocated object is not needed anymore is unknown. Then be conservative
// and use Heap::cleanup "late". I used this class with yacc and lex.
template<class T>
class Heap {
private:
public:
  set<T*> heap;
  const string type;

public:

  static const bool SILENT;
  static const bool VERBOSE;

  inline Heap(const string& type = ""): type(type) {}

  inline T*   request() { T* const x = new T; assert (insert(x)); return x; }
  inline bool insert(T* const x) { return heap.insert(x).second; }
  inline bool erase (T* const x) { return heap.erase(x); }

  void cleanup(const bool& verbose = SILENT); };

template<class T> const bool Heap<T>::SILENT  = false;
template<class T> const bool Heap<T>::VERBOSE = true;

template <class T>
void Heap<T>::cleanup(const bool& verbose) {
  if (verbose)
    std::cerr << "Heap<" << type << ">::cleanup: deleting " << heap.size() << " "
	      << ( type.empty() ? type : type + " " ) << "pointer" << ( heap.size() == 1 ? "" : "s" ) << std::endl;
  for (typename set<T*>::iterator p = heap.begin(); p != heap.end(); ++p)
    delete *p;
  heap.clear(); }

#endif
