#ifndef __MANAGER__
#define __MANAGER__

#include <string>
#include <iostream>

#include "functions.hh"
#include "statistics.hh"

#include "utool.hh"
#include "stateSpace.hh"


extern UCudd* cudd;


// Class that collects statistics information about all BDDs in existence,
// helps to print BDDs, and reports results on individual computed BDDs.

class Manager {
private:
  Statistics stats;
  const StateSpace& S;

public:
  Print_curr_S      print_curr_S;
  Print_next_S      print_next_S;
  Print_curr_next_S print_curr_next_S;

  Manager(const StateSpace& S): S(S), print_curr_S(S), print_next_S(S), print_curr_next_S(S) {}

  inline void start(const string& name)             { stats.start(name); }
  inline void end  (const UBDD& bdd = UBDD::Zero()) { stats.end  (bdd); }


  void print                  (                    cushort& indent = 0, ostream& out = cout) const;
  void print                  (const UBDD& result, cushort& indent = 0, ostream& out = cout); // non-const since passes Print_curr_S&, etc
  void print_decimal          (const UBDD& result,                      ostream& out = cout) const;
  void print_peakLiveNodeCount(                    cushort& indent = 0, ostream& out = cout) const; };

#endif
