#include <stdlib.h>

#include <list>
#include <fstream>

#include "utool.hh"
#include "model-check.hh"

// performs forward search
void ModelCheck::INV_FS(const UBDD& Good) {
  List<UBDD> images;
  UBDD Init = M.AP("Init"), Bad = ! Good, Unexplored = Init, Reached = Unexplored;
  if (! test_and_update_FS(Unexplored, Bad, images)) {
    counterExample_FS(images);
    return; }
  Report r("INV_FS", verbose);
  while (! Unexplored.empty()) {
    UBDD Successors = EY(Unexplored);
    Unexplored = Successors & (! Reached);
    if (! test_and_update_FS(Unexplored, Bad, images)) {
      counterExample_FS(images);
      return; }
    Reached |= Unexplored;
    r.print(Unexplored, Reached); }

  cout << "  -- INVARIANT HOLDS" << endl; }

// This function can be used no matter whether symmetry is turned on or not
bool ModelCheck::test_and_update_FS(const UBDD& Unexplored, const UBDD& Bad, List<UBDD>& images) const {
  UBDD errors = Unexplored & Bad;
  return ( errors.empty() ?
	   ( images.push_back(Unexplored), true  ) :    // test passed
	   ( images.push_back(errors),     false ) ); } // test failed

// the last UBDD in the images list is the set of reached Bad states
void ModelCheck::counterExample_FS(const List<UBDD>& images) const {
  cout << "  *** INVARIANT FAILED; generating counter example trace of edge length " << images.size() - 1 << " ... "; cout.flush();
  __SAFE_ASSERT__ (! images.empty());
  UBDD error = images.back().randomCube(0, UCudd::curr);
  List<UBDD> path;
  path.push_front(error);
  for (List<UBDD>::const_reverse_iterator image = images.succ(images.rbegin()); image != images.rend(); ++image) { // note: starts at 2nd element
    error = (EX(error) & (*image)).randomCube(0, UCudd::curr);
    path.push_front(error); }
  print(path); }

// performs backward search
void ModelCheck::INV_BS(const UBDD& Good) {
  List<UBDD> images;
  UBDD Init = M.AP("Init"), Bad = ! Good, Unexplored = Bad, Reached = Unexplored;
  if (! test_and_update_BS(Unexplored, Init, images)) {
    counterExample_BS(images);
    return; }
  Report r("INV_BS", verbose);
  while (! Unexplored.empty()) {
    UBDD Successors = EX(Unexplored);
    Unexplored = Successors & (! Reached);
    if (! test_and_update_BS(Unexplored, Init, images)) {
      counterExample_BS(images);
      return; }
    Reached |= Unexplored;
    r.print(Unexplored, Reached); }

  cout << "  -- INVARIANT HOLDS" << endl; }

// This function can be used no matter whether symmetry is turned on or not
bool ModelCheck::test_and_update_BS(const UBDD& Unexplored, const UBDD& Init, List<UBDD>& images) const {
  UBDD errors = Unexplored & Init;
  return ( errors.empty() ?
	   ( images.push_front(Unexplored), true  ) :    // test passed
	   ( images.push_front(errors),     false ) ); } // test failed

// the first UBDD in the images list is the set of backward-reached Init states.
// This routine could also print the evolving trace right away, instead of calling print
void ModelCheck::counterExample_BS(const List<UBDD>& images) const {
  cout << "  *** INVARIANT FAILED; generating counter example trace of edge length " << images.size() - 1 << " ... "; cout.flush();
  __SAFE_ASSERT__ (! images.empty());
  UBDD error = images.front().randomCube(0, UCudd::curr);
  List<UBDD> path;
  path.push_back(error);
  for (List<UBDD>::const_iterator image = images.succ(images.begin()); image != images.end(); ++image) { // note: starts at 2nd element
    error = (EY(error) & (*image)).randomCube(0, UCudd::curr);
    path.push_back(error); }
  print(path); }

// Uses a system call
void ModelCheck::print(const List<UBDD>& path, ostream& out) const {
  out << "done" << endl;
  if (CONTROL::wait_yes_no("Print counter example trace?")) {
    ushort s = 0;
    for (List<UBDD>::const_iterator state = path.begin(); state != path.end(); ++state, ++s) {
      int* cube = state->UBDD2Cube();
      out << endl;
      if (s == 0) {
	std::ofstream prev("/tmp/prev");
	out << "Initial state:" << endl;
	S.print_decimal(cube, UCudd::curr, out);
	S.print_decimal(cube, UCudd::curr, prev);
	CONTROL::wait_return(); }
      else {
	std::ofstream curr("/tmp/curr");
	out << "State " << s-1 << " (left) and state " << s << " (right), differences marked by '|':" << endl;
	S.print_decimal(cube, UCudd::curr, curr);
	curr.close();
	system("diff --side-by-side /tmp/prev /tmp/curr");
	system("mv -f /tmp/curr /tmp/prev");
	CONTROL::wait_return(); }}
    out << endl; }}

void ModelCheck::test() {}
