#ifndef __MODELCHECK__
#define __MODELCHECK__

#include <string>
#include <iostream>

#include "functions.hh"
#include "utool.hh"
#include "stateSpace.hh"
#include "manager.hh"
#include "model.hh"


class ModelCheck {
protected:
  const StateSpace& S;
  Manager& Man;
  Model& M;
  const UBDD&  States; // curr-states
  const UBDD& _States; // next-states
  const UBDD& R;

  const bool& verbose;

public:
  ModelCheck(const StateSpace& S, Manager& Man, Model& M, const bool& verbose): S(S), Man(Man), M(M), States(S.States), _States(S._States), R(M.R), verbose(verbose) {}
  virtual inline ~ModelCheck() {}

  inline void start (const string& name)                      const { Man.start (name); }
  inline void end   (const UBDD& bdd = UBDD::Zero())          const { Man.end   (bdd); } // default value given for INV routines (which don't return anything)
  inline void print (const UBDD& result, cushort& indent = 0) const { Man.print(result, indent); }
  inline UBDD the_States()                                    const { return States; }
  inline void enumerate_R()                                   const { R.enumerate(Man.print_curr_next_S, UBDD::lines_per_page); }

  virtual void test();

public:
  // the following are overloaded in ModelCheck_reduced
  virtual inline UBDD neg  (const UBDD& Z)       { return !Z; }
  virtual inline UBDD alpha(const UBDD& Z) const { throw CONTROL::Error("ModelCheck::alpha: undefined for concrete structure"); }
  virtual inline UBDD Alpha(const UBDD& Z) const { throw CONTROL::Error("ModelCheck::Alpha: undefined for concrete structure"); }
  virtual inline UBDD orbit(const UBDD& Z) const { throw CONTROL::Error("ModelCheck::orbit: undefined for concrete structure"); }

  inline UBDD EX(const UBDD& Z) const { return R.EX(Z); }
  inline UBDD AX(const UBDD& Z) const { return R.AX(Z); }
  inline UBDD EY(const UBDD& Z) const { return R.EY(Z); }
  inline UBDD AY(const UBDD& Z) const { return R.AY(Z); }

  virtual inline UBDD EF    (const UBDD& f)                { return R.EF    (f,    verbose); }
  virtual inline UBDD AF    (const UBDD& f)                { return R.AF    (f,    verbose); }
  virtual inline UBDD EG    (const UBDD& f)                { return R.EG    (f,    verbose); }
  virtual inline UBDD AG    (const UBDD& f)                { return R.AG    (f,    verbose); }
  virtual inline UBDD EU    (const UBDD& f, const UBDD& g) { return R.EU    (f, g, verbose); }
  virtual inline UBDD AU    (const UBDD& f, const UBDD& g) { return R.AU    (f, g, verbose); }
  virtual inline UBDD EP    (const UBDD& f)                { return R.EP    (f,    verbose); }
  virtual inline UBDD EX_INF(const UBDD& f)                { return R.EX_INF(f,    verbose); }
  virtual inline UBDD EY_INF(const UBDD& f)                { return R.EY_INF(f,    verbose); }
  virtual inline UBDD FS    (const UBDD& f)                { return R.FS    (f,    verbose); }
  virtual inline UBDD BS    (const UBDD& f)                { return R.BS    (f,    verbose); }

  virtual void INV_FS(const UBDD& Good);
  virtual void INV_BS(const UBDD& Good);
  bool test_and_update_FS(const UBDD& Unexplored, const UBDD& Bad,  List<UBDD>& images) const;
  bool test_and_update_BS(const UBDD& Unexplored, const UBDD& Init, List<UBDD>& images) const;
  virtual void counterExample_FS(const List<UBDD>& images) const;
  virtual void counterExample_BS(const List<UBDD>& images) const;
  void print(const List<UBDD>& path, ostream& out = cout) const; };

#endif
