#include <assert.h>
#include <iostream>
#include <string>
#include <map>
#include <vector>
#include <algorithm>

#include "functions.hh"
#include "triple.hh"

#include "stateSpace.hh"
#include "utool.hh"


extern UCudd* ucudd;


void Symmblock::print(cushort& indent, ostream& out) const {
  __SAFE_ASSERT__ (n == local_idsts.size()); // sanity
  out << string(indent, ' ') << "offset     = " << offset     << endl
      << string(indent, ' ') << "n          = " << n          << endl
    // << string(indent, ' ') << "l          = " << l          << endl
      << string(indent, ' ') << "local_bits = " << local_bits << endl
      << endl
      << string(indent, ' ') << "Global idst variables pointing to this block (current-state):" << ( global_idsts.empty() ? " (none)" : "" ) << endl;
  for (Vector<Triple<ushort> >::const_iterator t = global_idsts.begin(); t != global_idsts.end(); ++t)
    out << string(indent + 2, ' ')
	<< "Triple: " << *t << ", bit vector: " << StateSpace::bits(t->first, t->second, t->third) << endl;
  out << string(indent, ' ') << "Local  idst variables pointing to this block (current-state):";
  if ( n == 0 ? true : local_idsts[0].empty() )
    out << " (none)" << endl;
  else {
    out << endl;
    for (ushort p = 0; p < n; ++p) {
      out << string(indent + 2, ' ') << "process " << p << ": ";
      for (Vector<Triple<ushort> >::const_iterator t = local_idsts[p].begin(); t != local_idsts[p].end(); ++t) {
	out << "Triple: " << *t << ", bit vector: "; StateSpace::bits(t->first, t->second, t->third).to_stream(out, " ") << endl; }}}}

const Vector<ushort>& StateSpace::bits(cushort& start, cushort& numbits, cushort& increment) {
  typedef map<Triple<ushort>,Vector<ushort> > Triple_Vec_Map;
  static Triple_Vec_Map bits;
  Triple<ushort> triple(start, numbits, increment);
  Triple_Vec_Map::const_iterator found = bits.find(triple);
  if (found == bits.end()) {
    Vector<ushort> result;
    result.reserve(numbits);
    ushort pos = start;
    for (ushort b = 0; b < numbits; ++b, pos += increment) // b is just a counter
      result.push_back(pos);
    std::pair<Triple_Vec_Map::const_iterator,bool> new_entry = bits.insert(Triple_Vec_Map::value_type(triple, result));
    __SAFE_ASSERT__ (new_entry.second);
    return new_entry.first->second; }
  else
    return found->second; }

const UBDD& StateSpace::cube(cushort& start, cushort& numbits, cushort& increment) {
  typedef map<Triple<ushort>,UBDD> Triple_UBDD_Map;
  static Triple_UBDD_Map cube;
  Triple<ushort> triple(start, numbits, increment);
  Triple_UBDD_Map::const_iterator found = cube.find(triple);
  if (found == cube.end()) {
    UBDD result = UBDD::cubify(bits(start, numbits, increment));
    std::pair<Triple_UBDD_Map::const_iterator,bool> new_entry = cube.insert(Triple_UBDD_Map::value_type(triple, result));
    __SAFE_ASSERT__ (new_entry.second);
    return new_entry.first->second; }
  else
    return found->second; }

void StateSpace::__print(cushort& indent, ostream& out) const {
  const string indent_s = string(indent, ' ');
  out << indent_s << "number of BDD variables: 2 * " << ucudd->ReadSize() / 2 << " = " << ucudd->ReadSize() << endl
      << indent_s << "__global_bits = " << __global_bits << endl
      << indent_s << "__N_total = " << __N_total << endl;
  for (ushort sbn = 0; sbn < __sbs.size(); ++sbn) {
    out << string(indent, ' ') << "Symmetry block " << sbn << ":" << endl;
    __sbs[sbn].print(indent + 2, out); }}

void StateSpace::print_binary(const int* const state, cushort& start, cushort& numbits, cushort& increment, ostream& out) const {
  for (ushort b = 0; b < numbits; ++b)
    out << bit2char(state[start + b * increment]); }

long StateSpace::decimal(const int* const state, cushort& start, cushort& numbits, cushort& increment) const {
  Vector<ushort> value_bits;
  value_bits.reserve(numbits);
  for (ushort b = 0; b < numbits; ++b)
    value_bits.push_back(state[start + b * increment]);
  return LOGARITHM::n_ary2decimal(value_bits, 2); }

// Library functions for the user

UBDD StateSpace::invariant(cushort& p, cushort& sbn) const {
  const Vector<ushort>& currs = bits(p, sbn, UCudd::curr);
  const Vector<ushort>& nexts = bits(p, sbn, UCudd::next);
  return UBDD::equal(currs, nexts); }

UBDD StateSpace::invariant_sbn(cushort& sbn) const {
  UBDD result = UBDD::One();
  for (int p = __sbs[sbn].n - 1; p >= 0; --p)
    result &= invariant(p, sbn);
  return result; }

UBDD StateSpace::invariant_but(cushort& p, cushort& sbn) const {
  UBDD result = UBDD::One();
  for (int other = __sbs[sbn].n - 1; other >= 0; --other)
    if (other != p)
      result &= invariant(other, sbn);
  return result; }

UBDD StateSpace::invariant_all() const {
  UBDD result = UBDD::One();
  for (int sbn = __sbs.size() - 1; sbn >= 0; --sbn)
    result &= invariant_sbn(sbn);
  return result; }

UBDD StateSpace::invariant_globals() const {
  UBDD result = UBDD::One();
  for (int b = __global_bits - 1; b >= 0; --b)
    result &= UBDD::equal(2 * b, 2 * b + 1);
  return result; }
