#include <string>
#include <list>
#include <stack>
#include <algorithm>

#include "functions.hh"
#include "statistics.hh"
#include "utool.hh"

using std::string;
using std::list;
using std::stack;
using std::max;

using namespace PPRINT;

extern UCudd* ucudd;

void Statistics::start(const string& name) {
  StatData s(name, util_cpu_time());
  stat_stack.push(s);
  if (verbose)
    cout << "started computing " << name << endl; }

void Statistics::end(const UBDD& B) {
  StatData s = stat_stack.top();
  if (verbose)
    cout << "finished computing " << s.name << endl;
  stat_stack.pop();
  s.computeTime = util_cpu_time() - s.computeTime;
  s.nodeCount = B.nodeCount();
  s.peakNodeCount = ucudd->ReadPeakLiveNodeCount();
  statistics.push_back(s); }

// Outputs the statistics information collected
void Statistics::print(cushort& indent, ostream& out) const {
  size_t nameLength = 0, nodeCountLength = 0, computeTimeLength = 0, peakNodeCountLength = 0;
  ulong totalComputeTime = 0;
  for (list<StatData>::const_iterator data = statistics.begin(); data != statistics.end(); ++data) {
    nameLength          = max(         nameLength, data->name.size());
    nodeCountLength     = max(    nodeCountLength, widthify(data->nodeCount).size());
    computeTimeLength   = max(  computeTimeLength, formatString(data->computeTime / 1000.0, "%.1f").size());
    peakNodeCountLength = max(peakNodeCountLength, widthify(data->peakNodeCount).size());
    totalComputeTime += data->computeTime; }

  const string i = string(indent, ' ');
  for (list<StatData>::const_iterator data = statistics.begin(); data != statistics.end(); ++data)
    out << i << widthify(data->name, nameLength, LEFTJUST) << ": "
	<< "node count = "           << tabularize(widthify(data->nodeCount, nodeCountLength, RIGHTJUST)) << ", "
	<< "comp. time (sec) = "     << widthify(formatString(data->computeTime / 1000.0, "%.1f"), computeTimeLength, RIGHTJUST) << ", "
	<< "peak live node count = " << tabularize(widthify(data->peakNodeCount, peakNodeCountLength, RIGHTJUST))
	<< endl;
  totalComputeTime /= 1000; // ms -> s
  out << i << "Total computation time: " << PPRINT::hourize(totalComputeTime) << endl; }
