#ifndef __STATISTICS_HH__
#define __STATISTICS_HH__

/* This class is used to collect space and time information about BDDs during
   their creation. Use like this:
   1. Create a variable: "Statistics S;"
   2. In a function that builds some BDD B that you want to collect this info about, say
      "S.start(name)" at the time/place where you want the time count to start
   3. When the BDD computation has finished, say "S.end(B)"
   4. At the end of your program, when you want to output the information, say
      "S.print(2);" or so ('2': prints 2 blanks before each line as indentation).
*/

#include <string>
#include <list>
#include <stack>

#include "functions.hh"
#include "utool.hh"

using std::string;
using std::list;
using std::stack;

class Statistics {

public:
  typedef enum { SILENT, VERBOSE } Verbose;

private:
  struct StatData {
    const string name;
    ulong computeTime; // also serves as storage for start time while computation is ongoing
    ulong nodeCount;
    ulong peakNodeCount;

    StatData(const string& name, culong& start_time): name(name), computeTime(start_time) {} };

  bool verbose;

  stack<StatData> stat_stack;
  list <StatData> statistics;

public:
  inline Statistics(const Verbose& verbose = VERBOSE): verbose(verbose == VERBOSE) {}
  void start(const string& name);
  void end(const UBDD& B = UBDD::Zero());
  inline void make_silent () { verbose = false; }
  inline void make_verbose() { verbose = true; }
  void print(cushort& indent = 0, ostream& out = cout) const; };

#endif
