
#include <stdlib.h>

#include <string>
#include <iostream>
#include <algorithm>

#include "functions.hh"
#include "symbint.hh"
#include "symtab.hh"

#include "sviss-symtab.hh"

string Entry::the_kind() const {
  switch(kind) {
  case PARAMETER: return "parameter";
  case CONSTANT:  return "constant";
  case VARIABLE:  return "variable";
  default:        throw CONTROL::Error("Entry::the_kind: internal: wrong case"); }}

void Parameter::print(cushort& indent, ostream& out) const {
  out << "parameter";
  if (! meaning      .empty()) out << " (meaning: "       << meaning       << ")";
  if (! default_value.empty()) out << " (default_value: " << default_value << ")"; };

string Variable::the_type() const {
  switch (type) {
  case ENUMERATE: return "enumerate";
  case RANGE:     return "range";
  case ARRAY:     return "array";
  case RECORD:    return "record";
  default:        throw CONTROL::Error("Variable::the_type: internal: wrong case"); }}

void Variable::print(cushort& indent, ostream& out) const {
  out << the_scope() << " variable" << ( sb_pointer >= 0 ? " (id-sensitive)" : "" ) << " of type " << the_type() << ", ";
  out << "offset " << offset << ", numbits " << numbits << ", sbn " << sbn << ", sb_pointer " << sb_pointer; }

void Variable::print_binary(ostream& s, const SymbInt& offset, cushort& indent) const {
  s << string(indent, ' ') << "print_binary(state, ";
  if (scope == GLOBAL) {
    const SymbInt start = offset * SymbInt(2) + SymbInt("( part == UCudd::curr ? 0 : 1 )");
    s << start << ", " << numbits << ", 2);" << endl; }
  else
    s << "bit(p, " << sbn << ", " << offset << ", part), " << numbits << ", __local_increment(" << sbn << "));" << endl; }

void Variable::print_decimal(ostream& s, const SymbInt& offset, cushort& indent, const string& name) const {
  const string indent_before_name = "\"" + string( ( scope == GLOBAL ? 2 : 4 ), ' ') + "\"";
  const string indent_after_name  = "\"" + string( ( scope == GLOBAL ? 2 : 0 ), ' ') + "\"";
  s << string(indent, ' ') << "out << "
    << indent_before_name << " << " << name << " << "
    << indent_after_name  << " << \" = \" << decimal(state, ";
  if (scope == GLOBAL) {
    const SymbInt start = offset * SymbInt(2) + SymbInt("( part == UCudd::curr ? 0 : 1 )");
    s << start << ", " << numbits << ", 2) << endl;" << endl; }
  else
    s << "bit(p, " << sbn << ", " << offset << ", part), " << numbits << ", __local_increment(" << sbn << ")) << endl;" << endl; }

List<string> Enumerate::bool_identifiers("false, true", ", ");

void Enumerate::print_decimal(ostream& s, const SymbInt& offset, cushort& indent, const string& name) const {
  const string indent_s = string(indent, ' ');
  const string indent_before_name = "\"" + string( ( scope == GLOBAL ? 2 : 4 ), ' ') + "\"";
  const string indent_after_name  = "\"" + string( ( scope == GLOBAL ? 2 : 0 ), ' ') + "\"";
  const string value_name = PPRINT::varname("value");
  s << indent_s << "out << "
    << indent_before_name << " << " << name << " << "
    << indent_after_name  << " << \" = \";" << endl
    << indent_s << "culong " << value_name << " = decimal(state, ";
  if (scope == GLOBAL) {
    const SymbInt start = offset * SymbInt(2) + SymbInt("( part == UCudd::curr ? 0 : 1 )");
    s << start << ", " << numbits << ", 2);" << endl; }
  else
    s << "bit(p, " << sbn << ", " << offset << ", part), " << numbits << ", __local_increment(" << sbn << "));" << endl;
  s << indent_s << "switch(" << value_name << ") {" << endl;
  ushort i = 0;
  for (list<string>::const_iterator value = values.begin(); value != values.end(); ++value, ++i)
    s << indent_s << "  case " << i << ": out << \"" << (*value) << "\"; break;" << endl;
  s << indent_s << "}" << endl
    << indent_s << "out << endl;" << endl; }

void Enumerate::make_States(ostream& s, cushort& indent, const string& name) const {
  s << string(indent, ' ') << "States &= UBDD::atMost(" << name << "(" << ( scope == GLOBAL ? "" : "p" ) << "), " << values.size() - 1 << ");" << endl; }

Range::Range(cshort& sbn, const SymbInt& sb_pointer):
  Variable(RANGE, sbn, int(sb_pointer)), left(0), right(SymbInt("__sbs[" + string(sb_pointer) + "].n") - SymbInt(1)) {
  if (! sb_pointer.is_pos_literal())
    throw CONTROL::Error("Range::Range: argument of \"SB(...)\" (pointer to symmetry block) must be compile-time constant; received " + string(sb_pointer));
  make_numbits(); }

void Range::print(cushort& indent, ostream& out) const {
  Variable::print(indent, out);
  out << ", range boundaries "
       << SymbInt::parentheses(left) << ".." << SymbInt::parentheses(right) << " (inclusive)"; }

void Range::make_States(ostream& s, cushort& indent, const string& name) const {
  s << string(indent, ' ') << "States &= UBDD::atMost(" << name << "(" << ( scope == GLOBAL ? "" : "p" ) << "), " << right << ");" << endl; }

// printing array contents: print basevar size times, with additional offset v * basevar-> numbits (v=0..size-1)
void Array::print_binary(ostream& s, const SymbInt& offset, cushort& indent) const {
  const string indent_s = string(indent, ' ');
  const string v_name = PPRINT::varname("v");
  s << indent_s << "for (ushort " << v_name << " = 0; " << v_name << " < " << size << "; " << v_name << "++) {" << endl
    << indent_s << "  if (" << v_name << " > 0)" << endl
    << indent_s << "    out << \"|\";" << endl;
  basevar->print_binary(s, offset + SymbInt(v_name) * basevar->numbits, indent + 2);
  s << indent_s << "}" << endl; }

void Array::print_decimal(ostream& s, const SymbInt& offset, cushort& indent, const string& name) const {
  const string indent_s = string(indent, ' ');
  const string v_name = PPRINT::varname("v");
  s << indent_s << "for (ushort " << v_name << " = 0; " << v_name << " < " << size << "; " << v_name << "++) {" << endl;
  basevar->print_decimal(s, offset + SymbInt(v_name) * basevar->numbits, indent + 2, name + " << \"[\" << " + v_name + " << \"]\"");
  s << indent_s << "}" << endl; }

void Array::make_States(ostream& s, cushort& indent, const string& name) const {
  SymbInt base_size;
  switch (basevar->type) {
  case ENUMERATE: base_size = ((Enumerate*)basevar)->values.size() - 1; break;
  case RANGE:     base_size = ((Range*)basevar)->right; break;
  default:        CONTROL::warning("Array::make_States: 'States' only implemented for array basetypes enum and range"); }
  s << string(indent, ' ') << "for (ushort i = 0; i < " << size << "; ++i)" << endl
    << string(indent, ' ') << "  States &= UBDD::atMost(" << name << "(" << ( scope == GLOBAL ? "" : "p, " ) << "i), " << base_size << ");" << endl; }

void Record::make_offset(const SymbInt& offset) {
  Variable::make_offset(offset);
  SymbInt record_offset = 0;
  for (Symtab<Variable*>::iterator name_entry_pair = symtab.contents.begin(); name_entry_pair != symtab.contents.end(); ++name_entry_pair) {
    Variable*& var = name_entry_pair->entry();
    var->make_offset(offset + record_offset);
    record_offset = record_offset + var->numbits; }}

void Record::make_numbits() {
  numbits = 0;
  for (Symtab<Variable*>::const_iterator name_entry_pair = symtab.contents.begin(); name_entry_pair != symtab.contents.end(); ++name_entry_pair)
    numbits = numbits + name_entry_pair->entry()->numbits; }

void Record::print(cushort& indent, ostream& out) const {
  Variable::print(indent, out); out << ", components:" << endl;
  symtab.print(indent + 2, out); }

// printing record contents: print each variable, with offset increasing by numbits of each variable
void Record::print_binary(ostream& s, const SymbInt& offset, cushort& indent) const {
  SymbInt numbits = 0;
  for (Symtab<Variable*>::const_iterator name_entry_pair = symtab.contents.begin(); name_entry_pair != symtab.contents.end(); ++name_entry_pair) {
    if (name_entry_pair != symtab.contents.begin())
      s << string(indent, ' ') << "out << \"|\";" << endl;
    Variable* const& var = name_entry_pair->entry();
    var->print_binary(s, offset + numbits, indent);
    numbits = numbits + var->numbits; }}

void Record::print_decimal(ostream& s, const SymbInt& offset, cushort& indent, const string& name) const {
  SymbInt numbits = 0;
  for (Symtab<Variable*>::const_iterator name_entry_pair = symtab.contents.begin(); name_entry_pair != symtab.contents.end(); ++name_entry_pair) {
    Variable* const& var = name_entry_pair->entry();
    var->print_decimal(s, offset + numbits, indent, name + " << \"." + name_entry_pair->symname() + "\"");
    numbits = numbits + var->numbits; }}

void Record::make_States(ostream& s, cushort& indent, const string& name) const {
  CONTROL::warning("Record::make_States: 'States' not implemented for record types"); }

Record::~Record() {
  for (Symtab<Variable*>::const_iterator name_entry_pair = symtab.contents.begin(); name_entry_pair != symtab.contents.end(); ++name_entry_pair)
    delete name_entry_pair->entry(); }

string Sviss_Symtab::the_order() const {
  switch (order) {
  case CONCAT:  return "concatenated";
  case INTERL:  return "interleaved";
  case DYNAMIC: return "dynamic";
  default:      throw CONTROL::Error("Sviss_Symtab::the_order: internal: wrong case"); }}

void Sviss_Symtab::print(cushort& indent, ostream& out) const {
  Symtab<Entry*>::print(indent, out);
  out << endl;
  out << "Variable ordering: " << the_order() << endl; }

Sviss_Symtab::~Sviss_Symtab() {
  for (Symtab<Entry*>::const_iterator name_entry_pair = contents.begin(); name_entry_pair != contents.end(); ++name_entry_pair)
    delete name_entry_pair->entry(); }
