
%{

// Tokenizing a SVISS description

#include <stdlib.h>

#include <iostream>
#include <list>

using std::string; // needed in automatically created sviss.y.h
using std::list;   // ditto

#include "sviss-symtab.hh"
#include "sviss.y.h"
#include "parse.hh"

#define YY_DECL int yylex(Parse& parse)

%}

%x COMMENT
%x CODE

digit      [0-9]
number     {digit}+
letter     [A-Za-z]
identifier {letter}({letter}|{digit}|_)*
text       \"[^"\n]*\"
whitespace [ \t]
newline    \n


%%

{whitespace}+             ;
{newline}                 { ++parse.line_number; }

"//".*                    ; /* comment */
"/*"                      { BEGIN COMMENT; }
<COMMENT>{newline}        { ++parse.line_number; }
<COMMENT>"*/"             { BEGIN INITIAL; }
<COMMENT>.                ;

<CODE>.                   { parse.R_APs_cc << yytext[0]; }
<CODE>{newline}           { parse.R_APs_cc << yytext[0]; }

%{ // top level %}
"Parameter"               { return PARAMETER_TOK; }
"Const"                   { return CONST_TOK; }
"Order"                   { return ORDER_TOK; }
"Global"                  { return GLOBAL_TOK; }
"Clique"                  { return CLIQUE_TOK; }
"Ring"                    { return RING_TOK; }
"Proposition"             { return AP_TOK; }
"Code"                    { BEGIN CODE; return CODE_TOK; }

%{ // variable order %}
"concat"                  { return CONCAT_TOK; }
"interl"                  { return INTERL_TOK; }
"dynamic"                 { return DYNAMIC_TOK; }

%{ // type %}
"bool"                    { return BOOL_TOK; }
"enum"                    { return ENUM_TOK; }
"range"                   { return RANGE_TOK; }
"array"                   { return ARRAY_TOK; }
"of"                      { return OF_TOK; }
"record"                  { return RECORD_TOK; }

%{ // floating point function %}
"ceil_log"                { return CEIL_LOG_TOK; }
"power"                   { return POWER_TOK; }

%{ // misc %}
"SB"                      { return SB_TOK; }

%{ // character combination %}
".."                      { return DOTDOT_TOK; }

%{ // these must come last, as they otherwise clobber patterns like ceil_log %}
{number}                  { yylval.num = atoi(yytext); return NUM_TOK; }
{identifier}              { yylval.str = new string(yytext, yyleng); parse.check_reserved_words(*yylval.str); parse.s_heap.insert(yylval.str); return   ID_TOK; }
{text}                    { yylval.str = new string(yytext, yyleng);                                          parse.s_heap.insert(yylval.str); return TEXT_TOK; }

%{ // all else %}
.                         { return yytext[0]; }

%%

// To avoid warning about unused yyunput:
int __YYUNPUT_DUMMY__() {
  yyunput(0, 0);
  return 0; }
