
#ifndef __TRIPLE_HH__
#define __TRIPLE_HH__

#include <istream>
#include <ostream>

template<class X, class Y = X, class Z = Y>
struct Triple {
public:
  X first;
  Y second;
  Z third;

  inline Triple() {}
  inline Triple(const X& first, const Y& second, const Z& third): first(first), second(second), third(third) {} };

template<class X, class Y, class Z>
inline bool operator==(const Triple<X,Y,Z>& a, const Triple<X,Y,Z>& b) {
  return a.first == b.first && a.second == b.second && a.third == b.third; }

template<class X, class Y, class Z>
bool operator<(const Triple<X,Y,Z>& a, const Triple<X,Y,Z>& b) {
  { const X& af = a.first;  const X& bf = b.first;  if (af != bf) return af < bf; }
  { const Y& as = a.second; const Y& bs = b.second; if (as != bs) return as < bs; }
  { const Z& at = a.third;  const Z& bt = b.third;  if (at != bt) return at < bt; }
  return false; }

template<class X, class Y, class Z>
std::istream& operator>>(std::istream& in, Triple<X,Y,Z>& t) {
  X first;
  Y second;
  Z third;
  in >> first >> second >> third;
  t.first  = first;
  t.second = second;
  t.third  = third;
  return in; }

template<class X, class Y, class Z>
inline std::ostream& operator<<(std::ostream& out, const Triple<X,Y,Z>& t) {
  out << t.first << " " << t.second << " " << t.third; return out; }

#endif
